import sys
import math
import numpy as np
from scipy.stats import entropy
import matplotlib.pyplot as plt

#------------------------------------------------------------------------------------------------------------
class NodoHuffman:
    def __init__(self, char=None, prob=None, izq=None, der=None):
        self.char = char
        self.prob = prob
        
        self.izq = izq
        self.der = der
        # Los valores izq y der corresponden a las ramas izquierda (A) y derecha (B) que se combinan en 
        # un nuevo nodo (*) en cada paso del algoritmo de Huffman
        #   *
        #  / \
        # A   B

        self.bitstring = ''

    def __lt__(self, other):
        return self.prob < other.prob

#------------------------------------------------------------------------------------------------------------
def file2bin(file):
    """
    Toma el nombre de un archivo y convierte su contenido en bytes a un array de 0s y 1s. 

    Parameters
    ----------
    txt : str
        El nombre del archivo a ser convertido a binario.

    Returns
    -------
    Un array de numpy con 0s y 1s que representan el string original en la codificación.

    """
    fo = open(file, 'rb')
    msgbytes = fo.read()
    fo.close()
    
    # para generar la secuencia de 0s y 1s lo hago en dos pasos:
    # 1) convierto el archivo en bytes codificado a un entero (cuidado con el 
    # orden de los bits, sigamos el de ASCII)
    binint = int.from_bytes(msgbytes, 'big')
    # 2) uso el método format para convertir ese entero en un string de 0s y 1s. 
    # Asegurémonos que tiene la cantidad de bits correcta y no borre los primeros 0s.
    binstr = format(binint,'0'+str(len(msgbytes)*8)+'b')

    # Finalmente lo paso a un array numérico
    msgbin = np.array([int(bit, 2) for bit in binstr])
    return msgbin

#------------------------------------------------------------------------------------------------------------
def pdf(msgbin, orden):
    """
    Función que toma un mensaje en forma de array de bits, y calcula la función de distribución 
    empírica del alfabeto extendido 'orden' bits. 

    Parameters
    ----------
    msgbin : array de numpy con 0s y 1s.
        El mensaje original.
    orden : int
        El orden por el que se va a extender el alfabeto. O sea, se agrupan de 
        a `orden` bits, y ese será el nuevo alfabeto.

    Returns
    -------
    pdf : Array de numpy
        La probabilidad estimada de ocurrencia de cada símbolo del alfabeto extendido.

    """
    
    # Como el largo del mensaje no tiene porqué ser múltiplo de `orden`, le agrego al principio algunos 0s 
    # para que sí lo sea.
    lmsg  = len(msgbin)
    resto = lmsg % orden
    if resto:
        msgbin = np.pad(msgbin, (orden-resto,0), 'constant')
        lmsg  = len(msgbin)
    
    
    # Los símbolos van a quedar representados por el entero en base 2 de la palabra de bits correspondiente. 
    # Genero entonces un array con estos símbolos mediante una multiplicación matricial. (Los primeros bits 
    # son los más significativos.)
    msg_reshaped = msgbin.reshape(int(lmsg/orden),orden)
    base = np.power(2,np.arange(orden-1,-1,-1))
    
    # Finalmente calculo el histograma, usando bins uniformes entre 0 y 2^orden-1.
    (pdf, _) = np.histogram(msg_reshaped@base, bins=np.arange(2**orden + 1), density=True)
    return pdf

#------------------------------------------------------------------------------------------------------------
def entropia(x):

    """
    Calcula la entropía de la variable aleatoria.

    Parameters
    ----------
    x : Numpy array 
        Es la distribucion de probabilidades
        
    """

    return entropy(x, base=2, axis=0)

#------------------------------------------------------------------------------------------------------------
def huffman(p):
    """
    Genera un código de Huffman para la pdf `p`. 

    Parameters
    ----------
    p : Numpy array 
        Las probabilidades de cada símbolo del alfabeto. Se supone que 
        cada símbolo salió de hacer una extensión de una fuente binaria, 
        por lo que debería tener largo 2^extension. No se verifica.

    Returns
    -------
    codigo : dict
        Un diccionario cuyos keys son el símbolo del alfabeto extendido, y 
        el valor es el código optimizado.

    """
    
    # Suponemos que las palabras vienen de un alfabeto binario extendido, por lo que el total de 
    # símbolos debería ser 2^largo.
    largo_palabras = int(np.log2(len(p)))
    
    # Genero los nodos iniciales, que tienen especificadas las probabilidades y la palabra original.
    nodos = [NodoHuffman(char=format(i,'b').zfill(largo_palabras), prob=p[i]) for i in range(len(p))]

    # Construimos el árbol de Huffman combinando los dos nodos con menor probabilidad en cada paso, 
    # hasta que solo queda un nodo, que es la raíz del árbol.  
    while len(nodos) > 1:
    
        # En cada iteración se ordenan los nodos por su probabilidad en orden descendente.
        nodos = sorted(nodos, key=lambda x: x.prob, reverse=True)
        
        # Los dos nodos con menor probabilidad se extraen de la lista y se combinan en un nuevo nodo.
        izq_nodo = nodos.pop()
        der_nodo = nodos.pop()
        nuevo_nodo = NodoHuffman(prob=izq_nodo.prob + der_nodo.prob, izq=izq_nodo, der=der_nodo)
        
        # Finalmente se etiquetan las ramas de donde provienen con "0" y "1".
        izq_nodo.bitstring = '0'
        der_nodo.bitstring = '1'
        nodos.append(nuevo_nodo)

    codigo = {}
    
    def recorrer(nodo, bitstring=''):
        # Recorre el árbol generado y obtiene el código de Huffman correspondiente a cada símbolo del alfabeto. 
        
        if nodo.char is not None:
            codigo[nodo.char] = bitstring
        else:
            recorrer(nodo.izq, bitstring + nodo.izq.bitstring)
            recorrer(nodo.der, bitstring + nodo.der.bitstring)

    recorrer(nodos[0])
    return codigo

#------------------------------------------------------------------------------------------------------------
def codifica(msgbin,codigo):
    """
    Toma un mensaje en binario y un código (que tiene que corresponder con una versión extendida del 
    alfabeto binario original), y genera el mensaje en su nueva codificación.

    Parameters
    ----------
    msgbin : Array numpy binario
        Un array de 0s y 1s con el mensaje original sin codificar.
    codigo : dict
        Un diccionario cuyos keys sean las palabras del alfabeto extendido y los values la palabra codificada.

    Returns
    -------
    msg_codificado : Array numpy de strings 
        El mensaje codificado.

    """
    
    # Calculo primero la extensión del código.
    l = len(codigo)
    orden = int(math.log2(l))

    # Como el largo del mensaje no tiene porqué ser múltiplo de 'orden', le agrego al principio algunos 0s 
    # para que sí lo sea.
    lmsg  = len(msgbin)
    resto = lmsg % orden
    if resto:
        msgbin = np.pad(msgbin, (orden-resto,0), 'constant')
        lmsg  = len(msgbin)
    
    # Los símbolos van a quedar representados por el entero en base 2 de la palabra de bits correspondiente. 
    # Genero entonces un array con estos símbolos mediante una multiplicación matricial. (Los primeros bits 
    # son los más significativos.)
    msg_reshaped = msgbin.reshape(int(lmsg/orden),orden)
    base = np.power(2,np.arange(orden-1,-1,-1))
    msg = msg_reshaped@base
    
    # Ahora busco cada palabra en el código
    msg_codificado = [codigo[format(palabra,'b').zfill(orden)] for palabra in msg]

    # Finalmente devuelvo un array en el formato original, con 0s y 1s
    return np.array(sum([list(w) for w in msg_codificado],[]),dtype=int)


#------------------------------------------------------------------------------------------------------------
if __name__ == "__main__":         
    
    ############################################################
    # cambiar el nombre del archivo por el que se analizará
    ############################################################
    archivotxt = 'adams.txt'

    # convierto el archivo a binario
    msgbin = file2bin(archivotxt)

    # para varias extensiones de la fuente calculo la entropía y codifico con Huffman
    for i in range(4):
        orden = 2**i
        epdf = pdf(msgbin,orden);
        print('**************************************************')
        print('Entropía por bit original de la fuente extendida ', str(orden),' veces: ',str(entropia(epdf)/orden))
        
        #####################################
        # TODO graficar la distribución del
        # alfabeto extendido
        ######################################
        
        # insertar código aquí
        
        #######
        
        # calculo el código de Huffman óptimo para esta distribución
        codigo = huffman(epdf)

        #####################################
        # TODO 2: mostrar el código ordenando 
        # por largo de la palabra codificada
        #####################################
        
        # insertar código aquí
        
        ###############
    
        # dado el código diseñado, codifico el mensaje original
        msgCodif = codifica(msgbin,codigo)
        # print(msgCodif)

        # verifiquemos que la entropía del mensaje codificado sea lo que debe ser.
        epdf_huf = pdf(msgCodif,1)
        print('Entropía de la fuente Huffman codificada: ', str(entropia(epdf_huf)))
        print('Tasa de Compresión 1-largo(msg_codif)/largo(msg): ', str(1-len(msgCodif)/len(msgbin)))
